package com.esk.deob;

import static org.objectweb.asm.ClassWriter.COMPUTE_MAXS;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.zip.*;

import jode.decompiler.Decompiler;

import org.objectweb.asm.*;
import org.objectweb.asm.tree.*;

import com.esk.deob.impl.*;
import com.esk.tree.*;
import com.esk.util.Map;

public class Main {
  public static void main(String[] args) throws Throwable {
    if(args.length == 3 && args[0].equals("o"))
      deobfuscate(args[1], args[2]);
    else if(args.length == 3 && args[0].equals("c"))
      decompile(args[1], args[2]);
    else {
      System.out.println("Usage:");
      System.out.println("to deobfuscate: java com.esk.deob.Main o <src> <dst>");
      System.out.println("to decompile: java com.esk.deob.Main c <src> <dst>");
    }
  }

  @SuppressWarnings("unchecked")
  public static void deobfuscate(String src, String dst) throws Throwable {
    File f = new File(src);
    URI uri;
    if(!f.exists())
      uri = new URL(src).toURI();
    else
      uri = f.toURI();
    URL url = new URL("jar:" + uri + "!/");
    System.out.println(url);
    JarURLConnection jurlc = (JarURLConnection)url.openConnection();
    ZipFile zf = jurlc.getJarFile();
    ClassNode[] cs = new ClassNode[zf.size()];
    Enumeration<? extends ZipEntry> e = zf.entries();
    int i = 0;
    while(e.hasMoreElements()) {
      ZipEntry ze = e.nextElement();
      if(!ze.getName().endsWith(".class"))
        continue;
      ClassNode cn = new ClassNode();
      ClassReader cr = new ClassReader(zf.getInputStream(ze));
      cr.accept(cn, ClassReader.SKIP_FRAMES);
      cs[i++] = cn;
    }
    ClassNode[] _cns = new ClassNode[i];
    System.arraycopy(cs, 0, _cns, 0, i);
    cs = _cns;
    System.out.println("Building trees...");
    Map<MethodNode, NodeTree> nts = new Map<MethodNode, NodeTree>();
    int ct = 0, cur = 0;
    for(ClassNode cn : cs) {
      int pc = (int)Math.round((++ct / (double)cs.length) * 100);
      if(pc % 5 == 0 && pc != cur)
        System.out.print((cur = pc) + "%..");
      for(MethodNode mn : (java.util.List<MethodNode>)cn.methods) {
        NodeTree nt = new NodeTree(mn);
        nt.build();
        nts.put(mn, nt);
      }
    }
    System.out.print("\nFinding control field... ");
    ControlFieldFinder cff = new ControlFieldFinder();
    for(ClassNode cn : cs) {
      if(!cff.accept(cn))
        continue;
      cff.visitClasses(cs);
      for(MethodNode mn : (java.util.List<MethodNode>)cn.methods) {
        if(!cff.accept(mn))
          continue;
        NodeTree t = nts.get(mn);
        t.accept(cff);
      }
    }
    FieldHandle cf = cff.cf;
    System.out.println(cf);
    for(Transform t : new Transform[] { new TryCatch(),
                                        new OpaquePredicate(cf),
                                        new FlowRedundancy(),
                                        new NotComparison(),
                                        //new AddSubtract(),
                                        new ObnoxiousShift(), }) {
      System.out.println("Applying " + t.getClass().getName());
      for(ClassNode cn : cs) {
        if(!t.accept(cn))
          continue;
        t.visitClasses(cs);
        for(MethodNode mn : (java.util.List<MethodNode>)cn.methods) {
          if(!t.accept(mn))
            continue;
          NodeTree tr = nts.get(mn);
          tr.accept(t);
        }
      }
      t.printResult();
    }
    System.out.println("Collapsing trees into " + dst + "...");
    JarOutputStream jos = new JarOutputStream(new FileOutputStream(dst));
    double len = nts.size();
    ct = 0;
    cur = 0;
    for(ClassNode c : cs) {
      for(MethodNode m : (java.util.List<MethodNode>)c.methods) {
        NodeTree nt = nts.get(m);
        InsnList il = m.instructions = new InsnList();
        for(AbstractInsnNode ai : nt.collapse())
          il.add(ai);
        int pc = (int)Math.round((++ct / len) * 100);
        if(pc % 5 == 0 && pc != cur)
          System.out.print((cur = pc) + "%..");
        nts.removeByKey(m);
      }
      ClassWriter cw = new ClassWriter(COMPUTE_MAXS);
      c.accept(cw);
      JarEntry je = new JarEntry(c.name + ".class");
      jos.putNextEntry(je);
      jos.write(cw.toByteArray());
      jos.closeEntry();
    }
    System.out.println();
    jos.close();
  }

  public static void decompile(String cp, String dst) throws IOException {
    System.out.println("\nDecompiling from " + cp + " to " + dst + "...");
    Decompiler d = new Decompiler();
    com.esk.util.List<String> failed = new com.esk.util.List<String>();
    d.setClassPath(cp);
    JarFile jf = new JarFile(new File(cp));
    for(Enumeration<JarEntry> e = jf.entries(); e.hasMoreElements(); ) {
      JarEntry je = e.nextElement();
      String cn = je.getName();
      if(je.isDirectory() || !cn.endsWith(".class"))
        continue;
      cn = cn.substring(0, cn.length() - 6);
      System.out.println(cn);
      String fn = dst;
      if(!fn.endsWith("/"))
        fn += '/';
      fn += cn;
      File sf = new File(fn + ".java");
      sf.getParentFile().mkdirs();
      cn = cn.replace('/', '.');
      try {
        FileWriter fw = new FileWriter(sf);
        d.decompile(cn, fw, null);
        fw.close();
      } catch(Throwable t) {
        failed.add(cn);
        sf.delete();
        FileOutputStream fos = new FileOutputStream(fn + ".class");
        fos.write(read(jf.getInputStream(je)));
        fos.close();
      }
    }
    System.out.println("Unable to decompile: " + failed);
    
  }

  public static byte[] read(InputStream is) throws IOException {
    byte[] bs = new byte[0];
    byte[] b = new byte[1024];
    for(int i = 0; (i = is.read(b)) != -1; ) {
      byte[] _bs = new byte[bs.length + i];
      System.arraycopy(bs, 0, _bs, 0, bs.length);
      System.arraycopy(b, 0, _bs, bs.length, i);
      bs = _bs;
    }
    return bs;
  }
}