package com.esk.tree;

import org.objectweb.asm.tree.JumpInsnNode;

import com.esk.util.Util;

public class JumpNode extends Node {
  private Target t;

  public JumpNode(NodeTree t, JumpInsnNode ji, int c, int p) {
    super(t, ji, c, p);
  }

  public void accept(Visitor v) {
    v.visitJump(this);
  }

  public void reverse() {
    int o = op();
    switch(o) {
      case IFLT:
        o = IFGT;
        break;
      case IFGE:
        o = IFLE;
        break;
      case IFGT:
        o = IFLT;
        break;
      case IFLE:
        o = IFGE;
        break;
      case IF_ICMPLT:
        o = IF_ICMPGT;
        break;
      case IF_ICMPGE:
        o = IF_ICMPLE;
        break;
      case IF_ICMPGT:
        o = IF_ICMPLT;
        break;
      case IF_ICMPLE:
        o = IF_ICMPGE;
        break;
    }
    ai = new JumpInsnNode(o, insn().label);
  }

  public void setInsn(JumpInsnNode ji) {
    ai = ji;
  }

  public void setTarget(Target t) {
    if(this.t != null)
      this.t.ts.remove(this);
    this.t = t;
    if(t != null) {
      t.ts.add(this);
      insn().label = t.insn();
    } else
      insn().label = null;
  }

  public JumpInsnNode insn() {
    return (JumpInsnNode)ai;
  }

  public Target target() {
    return t;
  }

  public Node resolve() {
    return t.resolve();
  }

  public String toString(int i) {
    StringBuilder sb = new StringBuilder();
    sb.append(Util.toString(ai));
    sb.append(' ');
    sb.append('<');
    sb.append(' ');
    sb.append(Util.toString(t.resolve().ai));
    for(Node n : this) {
      sb.append('\n');
      for(int j = 0; j < i; j++)
        sb.append(' ');
      sb.append(n.toString(i + 1));
    }
    return sb.toString();
  }
}