package com.esk.tree;

import static org.objectweb.asm.tree.AbstractInsnNode.*;

import org.objectweb.asm.*;
import org.objectweb.asm.tree.*;

import com.esk.tree.Node;
import com.esk.util.*;

public class NodeTree extends Node implements Opcodes {
  private MethodNode mn;
  private Node[] ns;
  public static final int[] PDS, CDS;
  private int pos;

  static {
    int[] cds = new int[202];
    String c = "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACCCCCCCCBCBCBAAA"
             + "AAAAAAAAAAAAAAAAADEDEDDDDBCBCDCDECCECECECECECECECECECEBCBCCDCD"
             + "CDCECECEABBBCCCBBBCCCBBBECCEEBBBBBBCCCCCCCCAAABBBCBCBAAAAAAAAA"
             + "AABBBBBBBBAABBAA";
    for(int i = 0; i < cds.length; i++)
      cds[i] = c.charAt(i) - 'A';
    int[] pds = new int[202];
    String p = "ABBBBBBBBCCBBBCCBBBAABCBCBAAAAAAAAAAAAAAAAAAAABCBCBBBBAAAAAAAA"
             + "AAAAAAAAAAAAAAAAAAAAAAAAAAACDEEFGCBCBCBCBCBCBCBCBCBCBCBCBCBCBC"
             + "BCBCBCBCACBCBBCBCCBCBBBBBBBBBAAAAAAAAAAAAAAABAAAAAAAAAAAAAAAAA"
             + "ABBBBBBBAAAAAAAA";
    for(int i = 0; i < pds.length; i++)
      pds[i] = p.charAt(i) - 'A';
    PDS = pds;
    CDS = cds;
  }

  public NodeTree(MethodNode mn) {
    super(null, null, -1, -1);
    this.mn = mn;
  }

  @SuppressWarnings("unchecked")
  public void build() {
    List<JumpNode> jns = new List<JumpNode>();
    AbstractInsnNode[] ais = mn.instructions.toArray();
    int len = ais.length;
    ns = new Node[len];
    for(int i = 0; i < len; i++) {
      AbstractInsnNode ai = ais[i];
      int t = ai.getType(), o = ai.getOpcode(), c = 0, p = 0;
      switch(t) {
        case INSN:
        case INT_INSN:
        case VAR_INSN:
        case TYPE_INSN:
        case JUMP_INSN:
        case TABLESWITCH_INSN:
        case LOOKUPSWITCH_INSN:
          c = CDS[o];
          p = PDS[o];
          break;
        case FIELD_INSN:
          char d = ((FieldInsnNode)ai).desc.charAt(0);
          switch(o) {
            case Opcodes.GETSTATIC:
              p = d == 'D' || d == 'J' ? 2 : 1;
              break;
            case Opcodes.GETFIELD:
              c = 1;
              p = d == 'J' || d == 'D' ? 2 : 1;
              break;
            case Opcodes.PUTSTATIC:
              c = d == 'D' || d == 'J' ? 2 : 1;
              break;
            case Opcodes.PUTFIELD:
              c = d == 'J' || d == 'D' ? 3 : 2;
              break;
          }
          break;
        case METHOD_INSN:
          int as = Type.getArgumentsAndReturnSizes(((MethodInsnNode)ai).desc);
          if(o == INVOKESTATIC || o == INVOKEDYNAMIC)
            c = (as >> 2) - 1;
          else
            c = as >> 2;;
          p = as & 0x03;
          break;
        case LDC_INSN:
          Object cst = ((LdcInsnNode)ai).cst;
          p = cst instanceof Long || cst instanceof Double ? 2 : 1;
          break;
        case MULTIANEWARRAY_INSN:
          c = ((MultiANewArrayInsnNode)ai).dims;
          p = 1;
          break;
        case LABEL:
        case IINC_INSN:
        case FRAME:
        case LINE:
          break;
        default:
          throw new RuntimeException();
      }
      Node n = null;
      if(ai.getType() == AbstractInsnNode.LDC_INSN) {
        Object cst = ((LdcInsnNode)ai).cst;
        if(cst instanceof Number)
          n = new NumberNode(this, ai, c, p);
      } else if(o == BIPUSH || o == SIPUSH ||
          (o >= ICONST_M1 && o <= DCONST_1))
        n = new NumberNode(this, ai, c, p);
      if(o >= IRETURN && o <= RETURN)
        n = new ReturnNode(this, ai, c, p);
      if(n == null)
        switch(ai.getType()) {
          case FIELD_INSN:
          case METHOD_INSN:
            n = new InvocationNode(this, ai, c, p);
            break;
          case JUMP_INSN:
            JumpNode jn = new JumpNode(this, (JumpInsnNode)ai, c, p);
            jns.add(jn);
            n = jn;
            break;
          default:
            n = new Node(this, ai, c, p);
            break;
        }
      ns[i] = n;
    }
    for(int i = 0; i < ns.length; i++) {
      Node n = ns[i];
      if(n.ai.getType() != LABEL)
        continue;
      Target t = new Target(this, n.ai, n.c, n.p);
      boolean b = false;
      for(JumpNode jn : jns)
        if(jn.insn().label.getLabel().info == ns[i].ai) {
          jn.setTarget(t);
          b = true;
        }
      if(b)
        ns[i] = t;
      else
        ns[i].ai = new LabelNode(((LabelNode)ns[i].ai).getLabel());
    }
    /*Node[] _ns = new Node[len];
    len = 0;
    for(Node n : ns) {
      if(n.ai.getType() == LABEL && !(n instanceof Target))
        continue;
      _ns[len++] = n;
    }
    ns = Arrays.copyOf(_ns, len);*/
    pos = len - 1;
    for(TryCatchBlockNode tcb
        : (java.util.List<TryCatchBlockNode>)mn.tryCatchBlocks) {
      AbstractInsnNode s = Util.nextReal(tcb.start);
      AbstractInsnNode e = Util.nextReal(tcb.end);
      AbstractInsnNode h = Util.nextReal(tcb.handler);
      for(Node n : ns) {
        if(n.ai == s)
          n.s = true;
        if(n.ai == e)
          n.e = true;
        if(n.ai == h)
          n.h = true;
      }
    }
    while(true) {
      Node n = iter();
      if(n == null)
        break;
      /*if(head() != null && head().op() == ATHROW)
        if(n.r != 1)
          System.out.println(n + " " + n.r);*/
      add(0, n);
    }
  }

  public void accept(Visitor v) {
    for(Node n : this)
      accept(v, n);
  }

  private void accept(Visitor v, Node n) {
    n.accept(v);
    for(Node _n : n)
      accept(v, _n);
  }

  public AbstractInsnNode[] collapse() {
    AbstractInsnNode[] ais = super.collapse();
    int i = ais.length > 1 && ais[ais.length - 2].getType() == LABEL ? 2 : 1;
    return java.util.Arrays.copyOf(ais, ais.length - i);
  }

  public MethodNode method() {
    return mn;
  }

  private Node iter() {
    if(pos < 0)
      return null;
    Node n = ns[pos--];
    if(n.c == 0)
      return n;
    final int _c = n.h ? 0 : n.c;
    int c = _c;
    while(c != 0) {
      Node _n = iter();
      int o = _n.op();
      if(o == GOTO || o == JSR) {
        if(((JumpNode)_n).target().resolve() == n)
          c = _c;
      } else if(o == MONITOREXIT && n.op() == ATHROW)
        _n.r = 1;
      n.add(0, _n);
      int d = c - _n.r;
      if(d < 0) {
        n.r += -d;
        _n.r = 0;
        break;
      }
      c -= _n.r;
      _n.r = 0;
    }
    return n;
  }

  public String toString() {
    StringBuilder sb = new StringBuilder();
    for(Node n : this) {
      sb.append(n);
      sb.append('\n');
    }
    return sb.toString().trim();
  }
}