package com.esk.tree;

import org.objectweb.asm.tree.*;

public class NumberNode extends Node {
  public NumberNode(NodeTree t, AbstractInsnNode ai, int c, int p) {
    super(t, ai, c, p);
  }

  public void flipSign() {
    int s = sign();
    if(s == 0)
      return;
    Number n = value();
    if(n instanceof Long)
      setValue(-n.longValue());
    else if(n instanceof Float)
      setValue(-n.floatValue());
    else if(n instanceof Double)
      setValue(-n.doubleValue());
    else
      setValue(-n.intValue());
  }

  public void setValue(Number n) {
    if(n instanceof Long) {
      long l = n.longValue();
      if(l == 0L || l == 1L)
        ai = new InsnNode(LCONST_0 + (int)l);
      else
        ai = new LdcInsnNode(n);
      return;
    }
    if(n instanceof Float) {
      float f = n.floatValue();
      if(f == 0F || f == 1F || f == 2F)
        ai = new InsnNode(FCONST_0 + (int)f);
      else
        ai = new LdcInsnNode(n);
      return;
    }
    if(n instanceof Double) {
      double d = n.doubleValue();
      if(d == 0D || d == 1D)
        ai = new InsnNode(DCONST_0 + (int)d);
      else
        ai = new LdcInsnNode(n);
      return;
    }
    int i = n.intValue();
    if(i >= -1 && i <= 5)
      ai = new InsnNode(ICONST_0 + i);
    else if(i >= Byte.MIN_VALUE && i <= Byte.MAX_VALUE)
      ai = new IntInsnNode(BIPUSH, i);
    else if(i >= Short.MIN_VALUE && i <= Short.MAX_VALUE)
      ai = new IntInsnNode(SIPUSH, i);
    else
      ai = new LdcInsnNode(n);
  }

  public Number value() {
    if(ai.getType() == AbstractInsnNode.LDC_INSN)
      return (Number)((LdcInsnNode)ai).cst;
    int o = op();
    if(o == BIPUSH || o == SIPUSH)
      return ((IntInsnNode)ai).operand;
    if(o >= ICONST_M1 && o <= ICONST_5)
      return o - ICONST_0;
    if(o == LCONST_0 || o == LCONST_1)
      return (long)(o - LCONST_0);
    if(o >= FCONST_0 && o <= FCONST_2)
      return (float)(o - FCONST_0);
    if(o == DCONST_0 || o == DCONST_1)
      return (double)(o - DCONST_0);
    return null;
  }

  public int sign() {
    Number n = value();
    if(n instanceof Long) {
      long l = n.longValue();
      if(l == 0L)
        return 0;
      return l > 0 ? 1 : -1;
    }
    if(n instanceof Float) {
      float f = n.floatValue();
      if(f == 0F)
        return 0;
      return f > 0 ? 1 : -1;
    }
    if(n instanceof Double) {
      double d = n.doubleValue();
      if(d == 0D)
        return 0;
      return d > 0 ? 1 : -1;
    }
    int i = n.intValue();
    if(i == 0)
      return 0;
    return i > 0 ? 1 : -1;
  }
}