package com.esk.util;

import java.util.*;

public class List<T> extends AbstractSequentialList<T>
    implements Iterable<T>, Cloneable {
  private Node h;

  public void add(List<T> l) {
    for(T t : l)
      add(t);
  }

  public void add(T... ts) {
    for(T t : ts)
      add(t);
  }

  public void pushHead(T h) {
    add(0, h);
  }

  public static <T> List<T> forCollection(Collection<T> c) {
    List<T> l = new List<T>();
    l.addAll(c);
    return l;
  }

  public List<T> clone() {
    List<T> l = new List<T>();
    l.addAll(this);
    return l;
  }

  public List<T> subList(int f) {
    return subList(f, size() - 1);
  }

  public List<T> subList(int f, int t) {
    return (List<T>)super.subList(t, f);
  }

  public ListIterator<T> listIterator(int index) {
    return new ListIteratorImpl(index);
  }

  public Iterator<T> iterator() {
    return listIterator();
  }

  public String toString() {
    if(size() == 0)
      return "[]";
    StringBuilder sb = new StringBuilder("[");
    for(T t : this) {
      sb.append(t);
      sb.append(", ");
    }
    sb.append(']');
    int len = sb.length();
    return sb.delete(len - 3, len - 1).toString();
  }

  @SuppressWarnings("unchecked")
  public T[] array(Class<T> c) {
    T[] ts = (T[])java.lang.reflect.Array.newInstance(c, size());
    int i = 0;
    for(T t : this)
      ts[i++] = t;
    return ts;
  }

  public T popHead() {
    return remove(0);
  }

  public T popTail() {
    return remove(size() - 1);
  }

  public T head() {
    return h == null ? null : h.t;
  }

  public T tail() {
    int len = size();
    return len < 1 ? null : get(len - 1);
  }

  public int size() {
    int i = 0;
    for(Node n = h; n != null; i++, n = n.n);
    return i;
  }

  public boolean has() {
    return h != null;
  }

  public boolean equals(Object o) {
    return this == o;
  }

  private class Node {
    Node p, n;
    T t;

    Node(T t) {
      this.t = t;
    }
  }

  protected class ListIteratorImpl implements ListIterator<T> {
    Node p, n, lr;
    int i;

    public ListIteratorImpl(int i) {
      p = null;
      n = h;
      this.i = i;
      for(int j = 0; j < i; p = n, n = n.n, j++);
    }

    public void add(T t) {
      Node ins = new Node(t);
      if(h == null) {
        h = n = ins;
        lr = null;
        return;
      }
      if(p != null) {
        p.n = ins;
        ins.p = p;
      }
      if(n != null) {
        if(n == h)
          h = ins;
        Node _n = n;
        n = ins;
        n.n = _n;
        _n.p = n;
      }
      lr = null;
      i++;
    }

    public void set(T t) {
      if(lr == null)
        throw new IllegalStateException();
      lr.t = t;
    }

    public void remove() {
      if(lr == null)
        throw new IllegalStateException();
      if(lr != h) {
        if(lr == p) {
          p = p.p;
          if(p.p != null)
            p.p.n = p;
          p.n = n;
          if(p.n != null)
            p.n.p = p;
        } else
          throw new RuntimeException();
      } else {
        h = lr.n;
        if(h != null)
          h.p = null;
      }
      lr = null;
      i--;
    }

    public T next() {
      if(!hasNext())
        throw new NoSuchElementException();
      Node _n = lr = n;
      p = n;
      n = n.n;
      i++;
      return _n.t;
    }

    public T previous() {
      if(!hasPrevious())
        throw new NoSuchElementException();
      Node _n = lr = p;
      p = p.p;
      n = p;
      i--;
      return _n.t;
    }

    public T lastReturned() {
      return lr == null ? null : lr.t;
    }

    public int nextIndex() {
      return i;
    }

    public int previousIndex() {
      return i - 1;
    }

    public boolean hasNext() {
      return n != null;
    }

    public boolean hasPrevious() {
      return p != null;
    }
  };
}